/*
Copyright 1995 Massachusetts Institute of Technology.  All Rights 
Reserved.

You are hereby granted a worldwide, irrevocable, paid-up, right and 
license to use, execute, display, modify, copy and distribute MIT's 
Modifications, provided that (i) you abide by the terms and conditions 
of your Transarc AFS License Agreement, and (ii) you do not use the name 
of MIT in any advertising or publicity without the prior written consent 
of MIT.  MIT disclaims all liability for your use of MIT's 
Modifications.  MIT's Modifications are provided "AS IS" WITHOUT 
WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO, 
ANY WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE OR 
NONINFRINGEMENT.

*/
/*
 * NetBSD specific vnodeops + other misc interface glue
 */
/* $Id: afs_nbsdops.c,v 1.1.4.2 1999/05/24 23:07:23 sommerfeld Exp $ */
#ifndef lint
static char vcid[] = "$Id: afs_nbsdops.c,v 1.1.4.2 1999/05/24 23:07:23 sommerfeld Exp $";
#endif /* lint */
/*
 * A bunch of code cribbed from NetBSD ufs_vnops.c, ffs_vnops.c, and
 * nfs_vnops.c which carry this copyright: 
 */
/*
 * Copyright (c) 1982, 1986, 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#ident "$Revision: 1.1.4.2 $"

#include <afs/param.h>	/* Should be always first */
#ifdef	AFS_NETBSD_ENV
#include <afs/sysincludes.h>		/* Standard vendor system headers */
#include <afs/afsincludes.h>		/* Afs-based standard headers */
#include <afs/afs_stats.h>		/* statistics */
#include <sys/malloc.h>
#include <sys/namei.h>
#if (NetBSD1_2 > 1 || NetBSD >= 199710)
#define USE_GENFS
#endif
#ifdef USE_GENFS
#include <miscfs/genfs/genfs.h>
#endif

#define M_AFSNODE M_TEMP-1		/* XXX */

int afs_nbsd_lookup(void *);
int afs_nbsd_create(void *);
int afs_nbsd_mknod(void *);
int afs_nbsd_open(void *);
int afs_nbsd_close(void *);
int afs_nbsd_access(void *);
int afs_nbsd_getattr(void *);
int afs_nbsd_setattr(void *);
int afs_nbsd_read(void *);
int afs_nbsd_write(void *);
int afs_nbsd_ioctl(void *);
#ifdef USE_GENFS
#define	afs_nbsd_poll		genfs_poll
#else
int afs_nbsd_select(void *);
#endif
#if __NetBSD_Version__ - 0 >= 103050000
#define	afs_nbsd_revoke		genfs_revoke
#endif
int afs_nbsd_mmap(void *);
int afs_nbsd_fsync(void *);
int afs_nbsd_seek(void *);
int afs_nbsd_remove(void *);
int afs_nbsd_link(void *);
int afs_nbsd_rename(void *);
int afs_nbsd_mkdir(void *);
int afs_nbsd_rmdir(void *);
int afs_nbsd_symlink(void *);
int afs_nbsd_readdir(void *);
int afs_nbsd_readlink(void *);
#ifdef USE_GENFS
#define	afs_nbsd_abortop	genfs_abortop
#else
int afs_nbsd_abortop(void *);
#endif
int afs_nbsd_inactive(void *);
int afs_nbsd_reclaim(void *);
int afs_nbsd_lock(void *);
int afs_nbsd_unlock(void *);
int afs_nbsd_bmap(void *);
int afs_nbsd_strategy(void *);
int afs_nbsd_print(void *);
int afs_nbsd_islocked(void *);
int afs_nbsd_pathconf(void *);
int afs_nbsd_advlock(void *);
int afs_nbsd_truncate(void *);
int afs_nbsd_update(void *);

#ifdef USE_GENFS
#define	afs_nbsd_opnotsupp	genfs_eopnotsupp
#else
#define afs_nbsd_opnotsupp \
	((int (*) __P((void *)))eopnotsupp)
#endif
#define afs_nbsd_valloc afs_nbsd_opnotsupp
#define afs_nbsd_vfree afs_nbsd_opnotsupp
#define afs_nbsd_blkatoff afs_nbsd_opnotsupp
#define afs_nbsd_reallocblks afs_nbsd_opnotsupp

/* Global vfs data structures for AFS. */
int (**afs_nbsd_vnodeop_p)();
struct vnodeopv_entry_desc afs_nbsd_vnodeop_entries[] = {
	{ &vop_default_desc, vn_default_error },
	{ &vop_lookup_desc, afs_nbsd_lookup },		/* lookup */
	{ &vop_create_desc, afs_nbsd_create },		/* create */
	{ &vop_mknod_desc, afs_nbsd_mknod },		/* mknod */
	{ &vop_open_desc, afs_nbsd_open },		/* open */
	{ &vop_close_desc, afs_nbsd_close },		/* close */
	{ &vop_access_desc, afs_nbsd_access },		/* access */
	{ &vop_getattr_desc, afs_nbsd_getattr },	/* getattr */
	{ &vop_setattr_desc, afs_nbsd_setattr },	/* setattr */
	{ &vop_read_desc, afs_nbsd_read },		/* read */
	{ &vop_write_desc, afs_nbsd_write },		/* write */
	{ &vop_ioctl_desc, afs_nbsd_ioctl },		/* XXX ioctl */
#ifdef USE_GENFS
	{ &vop_poll_desc, afs_nbsd_poll },		/* poll */
#else
	{ &vop_select_desc, afs_nbsd_select },		/* select */
#endif
#if __NetBSD_Version__ - 0 >= 103050000
	{ &vop_revoke_desc, afs_nbsd_revoke },		/* revoke */
#endif
	{ &vop_mmap_desc, afs_nbsd_mmap },		/* mmap */
	{ &vop_fsync_desc, afs_nbsd_fsync },		/* fsync */
	{ &vop_seek_desc, afs_nbsd_seek },		/* seek */
	{ &vop_remove_desc, afs_nbsd_remove },		/* remove */
	{ &vop_link_desc, afs_nbsd_link },		/* link */
	{ &vop_rename_desc, afs_nbsd_rename },		/* rename */
	{ &vop_mkdir_desc, afs_nbsd_mkdir },		/* mkdir */
	{ &vop_rmdir_desc, afs_nbsd_rmdir },		/* rmdir */
	{ &vop_symlink_desc, afs_nbsd_symlink },	/* symlink */
	{ &vop_readdir_desc, afs_nbsd_readdir },	/* readdir */
	{ &vop_readlink_desc, afs_nbsd_readlink },	/* readlink */
	{ &vop_abortop_desc, afs_nbsd_abortop },	/* abortop */
	{ &vop_inactive_desc, afs_nbsd_inactive },	/* inactive */
	{ &vop_reclaim_desc, afs_nbsd_reclaim },	/* reclaim */
	{ &vop_lock_desc, afs_nbsd_lock },		/* lock */
	{ &vop_unlock_desc, afs_nbsd_unlock },		/* unlock */
	{ &vop_bmap_desc, afs_nbsd_bmap },		/* bmap */
	{ &vop_strategy_desc, afs_nbsd_strategy },	/* strategy */
	{ &vop_print_desc, afs_nbsd_print },		/* print */
	{ &vop_islocked_desc, afs_nbsd_islocked },	/* islocked */
	{ &vop_pathconf_desc, afs_nbsd_pathconf },	/* pathconf */
	{ &vop_advlock_desc, afs_nbsd_advlock },	/* advlock */
	{ &vop_blkatoff_desc, afs_nbsd_blkatoff },	/* blkatoff */
	{ &vop_valloc_desc, afs_nbsd_valloc },		/* valloc */
	{ &vop_reallocblks_desc, afs_nbsd_reallocblks }, /* reallocblks */
	{ &vop_vfree_desc, afs_nbsd_vfree },		/* vfree */
	{ &vop_truncate_desc, afs_nbsd_truncate },	/* truncate */
	{ &vop_update_desc, afs_nbsd_update },		/* update */
	{ &vop_bwrite_desc, vn_bwrite },		/* bwrite */
	{ (struct vnodeop_desc*)NULL, (int(*)())NULL }
};
struct vnodeopv_desc afs_nbsd_vnodeop_opv_desc =
	{ &afs_nbsd_vnodeop_p, afs_nbsd_vnodeop_entries };

#define GETNAME()	\
    struct componentname *cnp = ap->a_cnp; \
    char *name; \
    MALLOC(name, char *, cnp->cn_namelen+1, M_TEMP, M_WAITOK); \
    bcopy(cnp->cn_nameptr, name, cnp->cn_namelen); \
    name[cnp->cn_namelen] = '\0'

#define DROPNAME() FREE(name, M_TEMP)
	
#define NBSD_WRITES_ALLOWED
#ifndef NBSD_WRITES_ALLOWED
int nbsd_writes_allowed = 0;
#endif

#undef vrele
#define vrele afs_nbsd_rele
#undef VREF
#define VREF afs_nbsd_ref

int
afs_nbsd_lookup(v)
	void *v;
{
	struct vop_lookup_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
	} */ *ap = v;
    int error;
    struct vcache *vcp;
    struct vnode *vp, *dvp;
    register int flags = ap->a_cnp->cn_flags;
    int lockparent;			/* 1 => lockparent flag is set */
    int wantparent;			/* 1 => wantparent or lockparent flag */

    GETNAME();
    lockparent = flags & LOCKPARENT;
    wantparent = flags & (LOCKPARENT|WANTPARENT);

    if (ap->a_dvp->v_type != VDIR) {
	*ap->a_vpp = 0;
	DROPNAME();
	return ENOTDIR;
    }
    dvp = ap->a_dvp;
    if (afs_debug & AFSDEB_VNLAYER && !(dvp->v_flag & VROOT))
	printf("nbsd_lookup dvp %x flags %x name %s cnt %d\n", dvp, flags, name, dvp->v_usecount);
    error = afs_lookup(VTOVC(dvp), name, &vcp, cnp->cn_cred);
    if (error) {
	if ((cnp->cn_nameiop == CREATE || cnp->cn_nameiop == RENAME) &&
	    (flags & ISLASTCN) && error == ENOENT)
	    error = EJUSTRETURN;
	if (cnp->cn_nameiop != LOOKUP && (flags & ISLASTCN))
	    cnp->cn_flags |= SAVENAME;
	DROPNAME();
	*ap->a_vpp = 0;
	return (error);
    }
    vp = VCTOV(vcp);			/* always get a node if no error */

    /* The parent directory comes in locked.  We unlock it on return
       unless the caller wants it left locked.
       we also always return the vnode locked. */

    if (vp == dvp) {
	/* they're the same; afs_lookup() already ref'ed the leaf.
	   It came in locked, so we don't need to ref OR lock it */
	if (afs_debug & AFSDEB_VNLAYER)
	    printf("ref'ed %x as .\n", dvp);
    } else {
#if __NetBSD_Version__ - 0 >= 103050000
	if ((flags & ISDOTDOT) || !lockparent || !(flags & ISLASTCN))
	    VOP_UNLOCK(dvp, 0);		/* done with parent. */
	vn_lock(vp, LK_EXCLUSIVE | LK_RETRY);
	if (lockparent && (flags & ISDOTDOT) && (flags & ISLASTCN))
	    vn_lock(dvp, LK_EXCLUSIVE | LK_RETRY);
#else
	if ((flags & ISDOTDOT) || !lockparent || !(flags & ISLASTCN))
	    VOP_UNLOCK(dvp);		/* done with parent. */
	VOP_LOCK(vp);
	if (lockparent && (flags & ISDOTDOT) && (flags & ISLASTCN))
	    VOP_LOCK(dvp);
#endif	
	if (afs_debug & AFSDEB_VNLAYER)
	    printf("locked ret %x from lookup\n", vp);
    }
    *ap->a_vpp = vp;

    if ((cnp->cn_nameiop == RENAME && wantparent && (flags & ISLASTCN) ||
	 (cnp->cn_nameiop != LOOKUP && (flags & ISLASTCN))))
	cnp->cn_flags |= SAVENAME;

    DROPNAME();
    if (afs_debug & AFSDEB_VNLAYER && !(dvp->v_flag & VROOT))
	printf("nbsd_lookup done dvp %x cnt %d\n", dvp, dvp->v_usecount);
    return error;
}

int
afs_nbsd_create(v)
	void *v;
{
	struct vop_create_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap = v;
    int error = 0;
    struct vcache *vcp;
    register struct vnode *dvp = ap->a_dvp;
    GETNAME();

    if (afs_debug & AFSDEB_VNLAYER)
	printf("nbsd_create dvp %x cnt %d\n", dvp, dvp->v_usecount);

    /* vnode layer handles excl/nonexcl */

#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error = afs_create(VTOVC(dvp), name, ap->a_vap, NONEXCL,
		       ap->a_vap->va_mode, &vcp,
		       cnp->cn_cred);
    if (error) {
	VOP_ABORTOP(dvp, cnp);
	vput(dvp);
	DROPNAME();
	return(error);
    }

    if (vcp) {
	*ap->a_vpp = VCTOV(vcp);
#if __NetBSD_Version__ - 0 >= 103050000
	vn_lock(VCTOV(vcp), LK_EXCLUSIVE | LK_RETRY);
#else
	VOP_LOCK(VCTOV(vcp));
#endif
    }
    else *ap->a_vpp = 0;

    if ((cnp->cn_flags & SAVESTART) == 0)
	FREE(cnp->cn_pnbuf, M_NAMEI);
    vput(dvp);
    DROPNAME();
    if (afs_debug & AFSDEB_VNLAYER)
	printf("nbsd_create done dvp %x cnt %d\n", dvp, dvp->v_usecount);
    return error;
}

int
afs_nbsd_mknod(v)
	void *v;
{
	struct vop_mknod_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap = v;

    free(ap->a_cnp->cn_pnbuf, M_NAMEI);
    vput(ap->a_dvp);
    return(ENODEV);
}

int
afs_nbsd_open(v)
	void *v;
{
	struct vop_open_args /* {
		struct vnode *a_vp;
		int a_mode;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    int error;
    struct vcache *vc = VTOVC(ap->a_vp);
    error = afs_open(&vc, ap->a_mode, ap->a_cred);
#ifdef DIAGNOSTIC
    if (VCTOV(vc) != ap->a_vp)
	panic("AFS open changed vnode!");
#endif
    return error;
}

int
afs_nbsd_close(v)
	void *v;
{
	struct vop_close_args /* {
		struct vnode *a_vp;
		int a_fflag;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    return afs_close(VTOVC(ap->a_vp), ap->a_fflag, ap->a_cred, ap->a_p);
}

int
afs_nbsd_access(v)
	void *v;
{
	struct vop_access_args /* {
		struct vnode *a_vp;
		int a_mode;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    return afs_access(VTOVC(ap->a_vp), ap->a_mode, ap->a_cred);
}
int
afs_nbsd_getattr(v)
	void *v;
{
	struct vop_getattr_args /* {
		struct vnode *a_vp;
		struct vattr *a_vap;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    return afs_getattr(VTOVC(ap->a_vp), ap->a_vap, ap->a_cred);
}
int
afs_nbsd_setattr(v)
	void *v;
{
	struct vop_setattr_args /* {
		struct vnode *a_vp;
		struct vattr *a_vap;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    return afs_setattr(VTOVC(ap->a_vp), ap->a_vap, ap->a_cred);
}
int
afs_nbsd_read(v)
	void *v;
{
	struct vop_read_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;

    return afs_read(VTOVC(ap->a_vp), ap->a_uio, ap->a_cred, 0, 0, 0);
}
int
afs_nbsd_write(v)
	void *v;
{
	struct vop_write_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		int a_ioflag;
		struct ucred *a_cred;
	} */ *ap = v;

#ifndef UVM
    vnode_pager_uncache(ap->a_vp);	/* toss stale pages */
#else
    uvm_vnp_uncache(ap->a_vp);	/* toss stale pages */
#endif
    return afs_write(VTOVC(ap->a_vp), ap->a_uio, ap->a_ioflag, ap->a_cred, 0);
}

int
afs_nbsd_ioctl(v)
	void *v;
{
	struct vop_ioctl_args /* {
		struct vnode *a_vp;
		u_long a_command;
		caddr_t a_data;
		int a_fflag;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;
    struct vcache *tvc = VTOVC(ap->a_vp);
    struct afs_ioctl data;
    int error = 0;
  
    /* in case we ever get in here... */

    AFS_STATCNT(afs_ioctl);
    if (((ap->a_command >> 8) & 0xff) == 'V') {
	/* This is a VICEIOCTL call */
	error = HandleIoctl(tvc, (struct file *)0/*Not used*/,
			    ap->a_command, ap->a_data);
	return(error);
    } else {
	/* No-op call; just return. */
	return(ENOTTY);
    }
}

#ifndef USE_GENFS
/* ARGSUSED */
int
afs_nbsd_select(v)
	void *v;
{
	struct vop_select_args /* {
		struct vnode *a_vp;
		int a_which;
		int a_fflags;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

	/*
	 * We should really check to see if I/O is possible.
	 */
	return (1);
}
#endif

/*
 * Mmap a file
 *
 * NB Currently unsupported.
 */
/* ARGSUSED */
int
afs_nbsd_mmap(v)
	void *v;
{
	struct vop_mmap_args /* {
		struct vnode *a_vp;
		int a_fflags;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

	return (EINVAL);
}

int
afs_nbsd_fsync(v)
	void *v;
{
	struct vop_fsync_args /* {
		struct vnode *a_vp;
		struct ucred *a_cred;
#if __NetBSD_Version__ - 0 >= 103060000
                int a_flags;
#else
                int a_waitfor;
#endif
		struct proc *a_p;
	} */ *ap = v;
#if __NetBSD_Version__ - 0 >= 103060000
    int wait = (ap->a_flags&FSYNC_WAIT)? MNT_WAIT: 0;
#else
    int wait = ap->a_waitfor == MNT_WAIT;
#endif
    register struct vnode *vp = ap->a_vp;
    vflushbuf(vp, wait);
    return afs_fsync(VTOVC(vp), ap->a_cred);
}

int
afs_nbsd_seek(v)
	void *v;
{
	struct vop_seek_args /* {
		struct vnode *a_vp;
		off_t a_oldoff;
		off_t a_newoff;
		struct ucred *a_cred;
	} */ *ap = v;

    if (ap->a_newoff > ULONG_MAX)	/* AFS doesn't support 64-bit offsets */
	return EINVAL;
    return (0);
}

int
afs_nbsd_remove(v)
	void *v;
{
	struct vop_remove_args /* {
		struct vnode *a_dvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap = v;
    int error = 0;
    register struct vnode *vp = ap->a_vp;
    register struct vnode *dvp = ap->a_dvp;

    GETNAME();
#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error =  afs_remove(VTOVC(dvp), name, cnp->cn_cred);
    if (dvp == vp)
	vrele(vp);
    else
	vput(vp);
    vput(dvp);
    FREE(cnp->cn_pnbuf, M_NAMEI);
    DROPNAME();
    return error;
}

int
afs_nbsd_link(v)
	void *v;
{
	struct vop_link_args /* {
		struct vnode *a_vp;
		struct vnode *a_tdvp;
		struct componentname *a_cnp;
	} */ *ap = v;
    int error = 0;
#if NetBSD <= 199511
    register struct vnode *dvp = ap->a_vp;
    register struct vnode *vp = ap->a_tdvp;
#else
    register struct vnode *dvp = ap->a_dvp;
    register struct vnode *vp = ap->a_vp;
#endif

    GETNAME();
    if (dvp->v_mount != vp->v_mount) {
	VOP_ABORTOP(vp, cnp);
	error = EXDEV;
	goto out;
    }
    if (vp->v_type == VDIR) {
	VOP_ABORTOP(vp, cnp);
	error = EISDIR;
	goto out;
    }
#if __NetBSD_Version__ - 0 >= 103050000
    if (error = vn_lock(vp, LK_EXCLUSIVE)) {
#else
    if (error = VOP_LOCK(vp)) {
#endif
	VOP_ABORTOP(dvp, cnp);
	goto out;
    }
#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error = afs_link(VTOVC(vp), VTOVC(dvp), name, cnp->cn_cred);
    FREE(cnp->cn_pnbuf, M_NAMEI);
    if (dvp != vp)
#if __NetBSD_Version__ - 0 >= 103050000
	VOP_UNLOCK(vp, 0);
#else
	VOP_UNLOCK(vp);
#endif
out:
    vput(dvp);
    DROPNAME();
    return error;
}

int
afs_nbsd_rename(v)
	void *v;
{
	struct vop_rename_args  /* {
		struct vnode *a_fdvp;
		struct vnode *a_fvp;
		struct componentname *a_fcnp;
		struct vnode *a_tdvp;
		struct vnode *a_tvp;
		struct componentname *a_tcnp;
	} */ *ap = v;
    int error = 0;
    struct componentname *fcnp = ap->a_fcnp;
    char *fname;
    struct componentname *tcnp = ap->a_tcnp;
    char *tname;
    struct vnode *tvp = ap->a_tvp;
    register struct vnode *tdvp = ap->a_tdvp;
    struct vnode *fvp = ap->a_fvp;
    register struct vnode *fdvp = ap->a_fdvp;

    /*
     * Check for cross-device rename.
     */
    if ((fvp->v_mount != tdvp->v_mount) ||
	(tvp && (fvp->v_mount != tvp->v_mount))) {
	error = EXDEV;
abortit:
	VOP_ABORTOP(tdvp, tcnp); /* XXX, why not in NFS? */
	if (tdvp == tvp)
	    vrele(tdvp);
	else
	    vput(tdvp);
	if (tvp)
	    vput(tvp);
	VOP_ABORTOP(fdvp, fcnp); /* XXX, why not in NFS? */
	vrele(fdvp);
	vrele(fvp);
	return (error);
    }
    /*
     * if fvp == tvp, we're just removing one name of a pair of
     * directory entries for the same element.  convert call into rename.
     ( (pinched from NetBSD 1.0's ufs_rename())
     */
    if (fvp == tvp) {
	if (fvp->v_type == VDIR) {
	    error = EINVAL;
	    goto abortit;
	}

	/* Release destination completely. */
	VOP_ABORTOP(tdvp, tcnp);
	vput(tdvp);
	vput(tvp);

	/* Delete source. */
	vrele(fdvp);
	vrele(fvp);
	fcnp->cn_flags &= ~MODMASK;
	fcnp->cn_flags |= LOCKPARENT | LOCKLEAF;
	if ((fcnp->cn_flags & SAVESTART) == 0)
	    panic("afs_rename: lost from startdir");
	fcnp->cn_nameiop = DELETE;
	(void) relookup(fdvp, &fvp, fcnp);
	return (VOP_REMOVE(fdvp, fvp, fcnp));
    }
#if __NetBSD_Version__ - 0 >= 103050000
    if (error = vn_lock(fvp, LK_EXCLUSIVE))
#else
    if (error = VOP_LOCK(fvp))
#endif
	goto abortit;

    MALLOC(fname, char *, fcnp->cn_namelen+1, M_TEMP, M_WAITOK);
    bcopy(fcnp->cn_nameptr, fname, fcnp->cn_namelen);
    fname[fcnp->cn_namelen] = '\0';
    MALLOC(tname, char *, tcnp->cn_namelen+1, M_TEMP, M_WAITOK);
    bcopy(tcnp->cn_nameptr, tname, tcnp->cn_namelen);
    tname[tcnp->cn_namelen] = '\0';


#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    /* XXX use "from" or "to" creds? NFS uses "to" creds */
    error = afs_rename(VTOVC(fdvp), fname, VTOVC(tdvp), tname, tcnp->cn_cred);

#if __NetBSD_Version__ - 0 >= 103050000
    VOP_UNLOCK(fvp, 0);
#else
    VOP_UNLOCK(fvp);
#endif
    FREE(fname, M_TEMP);
    FREE(tname, M_TEMP);
    if (error)
	goto abortit;			/* XXX */
    if (tdvp == tvp)
	vrele(tdvp);
    else
	vput(tdvp);
    if (tvp)
	vput(tvp);
    vrele(fdvp);
    vrele(fvp);
    return error;
}

int
afs_nbsd_mkdir(v)
	void *v;
{
	struct vop_mkdir_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
	} */ *ap = v;
    register struct vnode *dvp = ap->a_dvp;
    register struct vattr *vap = ap->a_vap;
    int error = 0;
    struct vcache *vcp;

    GETNAME();
#ifdef DIAGNOSTIC
    if ((cnp->cn_flags & HASBUF) == 0)
	panic("afs_nbsd_mkdir: no name");
#endif
#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error = afs_mkdir(VTOVC(dvp), name, vap, &vcp, cnp->cn_cred);
    if (error) {
	VOP_ABORTOP(dvp, cnp);
	vput(dvp);
	DROPNAME();
	return(error);
    }
    if (vcp) {
	*ap->a_vpp = VCTOV(vcp);
#if __NetBSD_Version__ - 0 >= 103050000
	vn_lock(VCTOV(vcp), LK_EXCLUSIVE | LK_RETRY);
#else
	VOP_LOCK(VCTOV(vcp));
#endif
    } else
	*ap->a_vpp = 0;
    DROPNAME();
    FREE(cnp->cn_pnbuf, M_NAMEI);
    vput(dvp);
    return error;
}

int
afs_nbsd_rmdir(v)
	void *v;
{
	struct vop_rmdir_args /* {
		struct vnode *a_dvp;
		struct vnode *a_vp;
		struct componentname *a_cnp;
	} */ *ap = v;
    int error = 0;
    register struct vnode *vp = ap->a_vp;
    register struct vnode *dvp = ap->a_dvp;

    GETNAME();
    if (dvp == vp) {
	vrele(dvp);
	vput(vp);
	FREE(cnp->cn_pnbuf, M_NAMEI);
	DROPNAME();
	return (EINVAL);
    }

#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error = afs_rmdir(VTOVC(dvp), name, cnp->cn_cred);
    DROPNAME();
    vput(dvp);
    vput(vp);
    return error;
}

int
afs_nbsd_symlink(v)
	void *v;
{
	struct vop_symlink_args /* {
		struct vnode *a_dvp;
		struct vnode **a_vpp;
		struct componentname *a_cnp;
		struct vattr *a_vap;
		char *a_target;
	} */ *ap = v;
    register struct vnode *dvp = ap->a_dvp;
    int error = 0;
    /* NFS ignores a_vpp; so do we. */

    GETNAME();
#ifndef NBSD_WRITES_ALLOWED
    if (!nbsd_writes_allowed)
	error = EROFS;
    if (!error)
#endif
    error = afs_symlink(VTOVC(dvp), name, ap->a_vap, ap->a_target,
			cnp->cn_cred);
    DROPNAME();
    FREE(cnp->cn_pnbuf, M_NAMEI);
    vput(dvp);
    return error;
}

int
afs_nbsd_readdir(v)
	void *v;
{
	struct vop_readdir_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
		int *a_eofflag;
		off_t (1.3)/u_long(1.2,1.1) *a_cookies;
		int ncookies;
	} */ *ap = v;

/*    printf("readdir %x cookies %x ncookies %d\n", ap->a_vp, ap->a_cookies,
	   ap->a_ncookies); */
    return afs_readdir(VTOVC(ap->a_vp), ap->a_uio, ap->a_cred,
		       ap->a_eofflag, ap->a_cookies, ap->a_ncookies);
}

int
afs_nbsd_readlink(v)
	void *v;
{
	struct vop_readlink_args /* {
		struct vnode *a_vp;
		struct uio *a_uio;
		struct ucred *a_cred;
	} */ *ap = v;

/*    printf("readlink %x\n", ap->a_vp);*/
    return afs_readlink(VTOVC(ap->a_vp), ap->a_uio, ap->a_cred);
}

#ifndef USE_GENFS
int
afs_nbsd_abortop(v)
	void *v;
{
	struct vop_abortop_args /* {
		struct vnode *a_dvp;
		struct componentname *a_cnp;
	} */ *ap = v;

	if ((ap->a_cnp->cn_flags & (HASBUF | SAVESTART)) == HASBUF)
		FREE(ap->a_cnp->cn_pnbuf, M_NAMEI);
	return (0);
}
#endif

extern int prtactive;

int
afs_nbsd_inactive(v)
	void *v;
{
	struct vop_inactive_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
    register struct vnode *vp = ap->a_vp;
    int error;

    if (prtactive && vp->v_usecount != 0)
	vprint("afs_nbsd_inactive(): pushing active", vp);

    error = afs_inactive(VTOVC(vp), 0);	/* decrs ref counts */
#if __NetBSD_Version__ - 0 >= 103050000
    VOP_UNLOCK(vp, 0);
#endif
    return error;
}

int
afs_nbsd_reclaim(v)
	void *v;
{
	struct vop_reclaim_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
    int error;
    register struct vnode *vp = ap->a_vp;

    cache_purge(vp);			/* just in case... */
#ifndef UVM
    vnode_pager_uncache(vp);
#else
    uvm_vnp_uncache(vp);
#endif

    error = afs_FlushVCache(VTOVC(vp)); /* tosses our stuff from vnode */
    if (!error && vp->v_data)
	panic("afs_reclaim: vnode not cleaned");
    return error;
}

int
afs_nbsd_lock(v)
	void *v;
{
	struct vop_lock_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
	register struct vnode *vp = ap->a_vp;
	register struct vcache *vc = VTOVC(vp);

#if __NetBSD_Version__ - 0 >= 103050000
	if (vp->v_tag == VT_NON)
		return (ENOENT);
	return (lockmgr(&vc->kernlock, ap->a_flags | LK_CANRECURSE,
	    &vp->v_interlock));
#else
	while (vp->v_flag & VXLOCK) {
		vp->v_flag |= VXWANT;
		tsleep((caddr_t)vp, PINOD, "nbsdlock", 0);
	}
	if (vp->v_tag == VT_NON)
		return (ENOENT);
	if (vc->rwlockpid == osi_curproc()->p_pid) {
	    vc->locktrips++;
	    return 0;
	}
	while (vc->lockflags & NBSD_LOCK_LOCKED) {
#ifdef DIAGNOSTIC
	    if (!pfind(vc->rwlockpid))
		panic("afs_nbsd_lock: lock holder %d dead", vc->rwlockpid);
#endif
	    vc->lockflags |= NBSD_LOCK_WANTED;
	    tsleep((caddr_t)&vc->lockflags, PINOD, "afsvlock", 0);
	}
	vc->lockflags |= NBSD_LOCK_LOCKED;
	vc->rwlockpid = osi_curproc()->p_pid;
	vc->locktrips = 0;
/*	printf("afs_nbsd_lock %x, %d\n", vc, curproc->p_pid);*/
	return (0);
#endif
}

int
afs_nbsd_unlock(v)
	void *v;
{
	struct vop_unlock_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
	register struct vnode *vp = ap->a_vp;
	register struct vcache *vc = VTOVC(vp);

#if __NetBSD_Version__ - 0 >= 103050000
	return (lockmgr(&vc->kernlock, ap->a_flags | LK_RELEASE,
	    &vp->v_interlock));
#else
/*    printf("afs_nbsd_unlock %x, %d\n", vc, curproc->p_pid);*/
#ifdef DIAGNOSTIC
    if (vc->rwlockpid != osi_curproc()->p_pid) {
	printf("vc=%x, lockpid=%d,curpid=%d\n", vc, vc->rwlockpid,
	       curproc->p_pid);
	panic("nbsd_unlock: wrong caller");
    }
#endif
    if (vc->locktrips > 0) {
	--vc->locktrips;
	return;
    }
    vc->rwlockpid = -1;
    vc->lockflags &= ~NBSD_LOCK_LOCKED;
    wakeup((caddr_t)&vc->lockflags);
    return 0;
#endif
}

int
afs_nbsd_bmap(v)
	void *v;
{
	struct vop_bmap_args /* {
		struct vnode *a_vp;
		daddr_t  a_bn;
		struct vnode **a_vpp;
		daddr_t *a_bnp;
		int *a_runp;
	} */ *ap = v;
    struct vcache *vcp;
    int error;

    error = afs_bmap(VTOVC(ap->a_vp), ap->a_bn, &vcp, ap->a_bnp);
    if (ap->a_vpp) {
	if (vcp)
	    *ap->a_vpp = VCTOV(vcp);
	else *ap->a_vpp = 0;
    }
    return 0;
}
int
afs_nbsd_strategy(v)
	void *v;
{
	struct vop_strategy_args /* {
		struct buf *a_bp;
	} */ *ap = v;

    return afs_strategy(ap->a_bp);
}
int
afs_nbsd_print(v)
	void *v;
{
	struct vop_print_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
    register struct vnode *vp = ap->a_vp;
    register struct vcache *vc = VTOVC(ap->a_vp);

    printf("tag %d, fid: %ld.%x.%x.%x,", vp->v_tag, vc->fid.Cell,
	   vc->fid.Fid.Volume, vc->fid.Fid.Vnode, vc->fid.Fid.Unique);
#if __NetBSD_Version__ - 0 >= 103050000
    lockmgr_printinfo(&vc->kernlock);
#else
    if (vc->lockflags & NBSD_LOCK_LOCKED) {
	printf(" locked %d times by %d", vc->locktrips, vc->rwlockpid);
	if (vc->lockflags & NBSD_LOCK_WANTED)
	    printf(" (wanted)");
    }
#endif
    printf("\n");
    return 0;
}

int
afs_nbsd_islocked(v)
	void *v;
{
	struct vop_islocked_args /* {
		struct vnode *a_vp;
	} */ *ap = v;
    struct vcache *vc = VTOVC(ap->a_vp);

#if __NetBSD_Version__ - 0 >= 103050000
    return (lockstatus(&vc->kernlock));
#else
    if (vc->lockflags & NBSD_LOCK_LOCKED) {
#ifdef DIAGNOSTIC
	printf("afs vnode %x wanted by pid %d\n", osi_curproc()->p_pid);
	vprint("nbsd_islocked", ap->a_vp);
#endif
	return (1);
    }
    return (0);
#endif
}

/*
 * Return POSIX pathconf information applicable to ufs filesystems.
 */
afs_nbsd_pathconf(v)
	void *v;
{
	struct vop_pathconf_args /* {
		struct vnode *a_vp;
		int a_name;
		int *a_retval;
	} */ *ap = v;

    AFS_STATCNT(afs_cntl);
    switch (ap->a_name) {
      case _PC_LINK_MAX:
	*ap->a_retval = LINK_MAX;
	break;
      case _PC_NAME_MAX:
	*ap->a_retval = NAME_MAX;
	break;
      case _PC_PATH_MAX:
	*ap->a_retval = PATH_MAX;
	break;
      case _PC_CHOWN_RESTRICTED:
	*ap->a_retval = 1;
	break;
      case _PC_NO_TRUNC:
	*ap->a_retval = 1;
	break;
      case _PC_PIPE_BUF:
	return EINVAL;
	break;
      default:
	return EINVAL;
    }
    return 0;
}

/*
 * Advisory record locking support (fcntl() POSIX style)
 */
int
afs_nbsd_advlock(v)
	void *v;
{
	struct vop_advlock_args /* {
		struct vnode *a_vp;
		caddr_t a_id;
		int a_op;
		struct flock *a_fl;
		int a_flags;
	} */ *ap = v;

    return afs_lockctl(VTOVC(ap->a_vp), ap->a_fl, ap->a_op, osi_curcred(),
		       (int) ap->a_id);
}

int
afs_nbsd_truncate(v)
	void *v;
{
	struct vop_truncate_args /* {
		struct vnode *a_vp;
		off_t a_length;
		int a_flags;
		struct ucred *a_cred;
		struct proc *a_p;
	} */ *ap = v;

    printf("stray afs_nbsd_truncate\n");
    return EOPNOTSUPP;
}

int
afs_nbsd_update(v)
	void *v;
{
	struct vop_update_args /* {
		struct vnode *a_vp;
		struct timeval *a_access;
		struct timeval *a_modify;
		int a_waitfor;
	} */ *ap = v;

    printf("stray afs_nbsd_update\n");
    return EOPNOTSUPP;
}
#endif /* AFS_NETBSD_ENV */
